import torch
import re
import math
from functools import partial
import matplotlib.pyplot as plt
from bijou.utils import ToolBox as tbox
from tqdm import tqdm


class CancelTrainException(Exception):
    pass


class CancelEpochException(Exception):
    pass


class CancelBatchException(Exception):
    pass


class StateErrorException(Exception):
    pass


def camel2snake(name):
    """
    生成callback的名字
    """
    _camel_re1 = re.compile('(.)([A-Z][a-z]+)')
    _camel_re2 = re.compile('([a-z0-9])([A-Z])')

    s1 = re.sub(_camel_re1, r'\1_\2', name)
    return re.sub(_camel_re2, r'\1_\2', s1).lower()


def dictformat(dic):
    return str({k: f'{v:0.6f}' for k, v in dic.items()})


class Callback():
    """
    Callback的基类
    """
    _order = 0

    def set_learner(self, learner):
        self.learner = learner

    def __getattr__(self, k):
        return getattr(self.learner, k)

    @property
    def name(self):
        name = re.sub(r'Callback$', '', self.__class__.__name__)
        return camel2snake(name or 'callback')

    def __call__(self, cb_name):
        f = getattr(self, cb_name, None)
        if f and f():
            return True
        return False


class AvgStats():
    def __init__(self, metrics, state):
        self.metrics, self.state = tbox.listify(metrics), state
        self.metric_names = ['loss'] + [m.__name__ for m in self.metrics]  # 构造metrics 名称

    def reset(self):
        self.tot_loss, self.count = 0., 0
        self.tot_mets = [0.] * len(self.metrics)

    @property
    def all_stats(self):
        return [self.tot_loss.item()] + self.tot_mets

    @property
    def avg_stats(self):
        return dict(zip(self.metric_names, [o/self.count for o in self.all_stats]))

    def __repr__(self):
        if not self.count:
            return ""
        return f"{self.state}: {self.avg_stats}"

    def accumulate(self, learner):
        bn = len(learner.xb)  # .shape[0]
        self.tot_loss += learner.loss * bn
        self.count += bn
        for i, m in enumerate(self.metrics):
            self.tot_mets[i] += m(learner.pred, learner.yb).item() * bn


class AvgStatsCallback(Callback):
    """
    该类型的回调最后只用一个，如果需要多个指标，则将多个指标组成列表，传到metrics参数即可
    """
    _order = 1

    def __init__(self, metrics, name='metrics'):
        self.train_stats = AvgStats(metrics, 'train')
        self.valid_stats = AvgStats(metrics, 'val')
        self.test_stats = AvgStats(metrics, 'test')
        self.name_ = name

    def begin_epoch(self):
        self.train_stats.reset()
        self.valid_stats.reset()

    def after_epoch(self):
        self.learner.metric_values = {'train': self.train_stats.avg_stats,
                                      'val': self.valid_stats.avg_stats}

    def begin_test(self):
        self.test_stats.reset()

    def after_loss(self):
        if self.state == 'train':
            stats = self.train_stats
        elif self.state == 'val':
            stats = self.valid_stats
        elif self.state == 'test':
            stats = self.test_stats
        else:
            raise StateErrorException('Learner STATE must be in ["train", "val", "test"]!')

        with torch.no_grad():
            stats.accumulate(self.learner)

    def after_batch(self):
        if self.state == 'train':
            self.learner.infos['train'] = self.train_stats.avg_stats
        elif self.state == 'val':
            self.learner.infos['val'] = self.valid_stats.avg_stats

    def after_test(self):
        print(self.test_stats)


class ProgressBarCallback(Callback):
    _order = 0
    train_str = ''
    val_str = ''

    def begin_epoch(self):
        batch_num = len(self.data.train_dl)
        self.pbar = tqdm(total=batch_num,
                         bar_format=f'E {self.n_epochs + 1:<4} ' + 'B {n_fmt} {l_bar} {rate_fmt} | {postfix[0]}',
                         unit=' batch', postfix=[''])

    def begin_validate(self):
        self.pbar.reset(total=len(self.data.valid_dl))

    def after_batch(self):
        if self.state == 'train':
            if self.learner.infos['train']:
                self.train_str = 'train-' + str(dictformat(self.infos['train']))
            self.learner.infos['train'] = []
        elif self.state == 'val':
            if self.learner.infos['val']:
                self.val_str = 'val-' + str(dictformat(self.infos['val']))
            self.learner.infos['val'] = []
        else:
            return

        if self.state == 'train':
            self.pbar.postfix[0] = self.train_str.replace("'", '')
            self.pbar.update()

        elif self.state == 'val':
            self.pbar.postfix[0] = self.train_str.replace("'", '') + '  ' + self.val_str.replace("'", '')
            self.pbar.update()

    def after_epoch(self):
        self.pbar.close()

class TrainEvalCallback(Callback):
    def begin_fit(self):
        self.learner.n_epochs = 0.
        self.learner.n_iter = 0

    def after_batch(self):
        if self.state != 'train':
            return
        self.learner.n_epochs += 1./self.iters
        self.learner.n_iter += 1

    def begin_epoch(self):
        self.learner.n_epochs = self.epoch
        self.model.train()
        self.learner.state = 'train'

    def begin_validate(self):
        self.model.eval()
        self.learner.state = 'val'


class TestCallback(Callback):
    _order = 2

    def begin_test(self):
        self.model.eval()
        self.learner.state = 'test'


class Recorder(Callback):
    metric_of_epochs = []

    def begin_fit(self):
        self.lrs = [[] for _ in self.opt.param_groups]
        self.losses = []

    def after_batch(self):
        if self.state != 'train':
            return
        for pg, lr in zip(self.opt.param_groups, self.lrs):
            lr.append(pg['lr'])
        self.losses.append(self.loss.detach().cpu())

    def after_epoch(self):
        self.metric_of_epochs.append(self.metric_values)
        self.learner.metric_values = {}

    def plot_lr(self, pgid=-1):
        plt.plot(self.lrs[pgid])

    def plot_loss(self, skip_last=0):
        plt.plot(self.losses[:len(self.losses)-skip_last], c='r')

    def plot_lr_loss(self, skip_last=0, pgid=-1):
        losses = [o.item() for o in self.losses]
        lrs = self.lrs[pgid]
        n = len(losses)-skip_last
        plt.xscale('log')
        plt.plot(lrs[:n], losses[:n])

    def plot_metrics(self):
        trains_vals = ld2dl(self.metric_of_epochs)
        trains = ld2dl(trains_vals['train'])
        vals = ld2dl(trains_vals['val'])
        plt.figure(figsize=[len(trains) * 5, 5])
        for i, m in enumerate(trains):
            plt.subplot(1, len(trains), i + 1)
            plt.plot(trains[m], label='train')
            plt.plot(vals[m], label='val')
            plt.title(m, y=-0.2)
            plt.legend()
        plt.subplots_adjust(bottom=0.2)


def ld2dl(ld):
    dl = {}
    for d in ld:
        for k, v in d.items():
            if k not in dl:
                dl[k] = []
            dl[k].append(v)
    return dl


class LR_Find(Callback):
    _order = 1

    def __init__(self, max_iter=100, min_lr=1e-6, max_lr=10):
        self.max_iter, self.min_lr, self.max_lr = max_iter, min_lr, max_lr
        self.best_loss = 1e9

    def begin_batch(self):
        if self.state != 'train':
            return
        pos = self.n_iter/self.max_iter
        lr = self.min_lr * (self.max_lr/self.min_lr) ** pos
        for pg in self.opt.param_groups:
            pg['lr'] = lr

    def after_step(self):
        if self.n_iter >= self.max_iter or self.loss > self.best_loss*10:
            raise CancelTrainException()
        if self.loss < self.best_loss:
            self.best_loss = self.loss


class CudaCallback(Callback):
    _order = 0

    def __init__(self, device):
        if device is None:
            device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        self.device = device

    def begin_fit(self):
        self.model.to(self.device)

    def begin_batch(self):
        self.learner.xb, self.learner.yb = self.xb.to(self.device), self.yb.to(self.device)

    def begin_predict(self):
        self.model.to(self.device)
        self.learner.pred_data_tensor = self.pred_data_tensor.to(self.device)


class BatchTransformXCallback(Callback):
    _order = 2

    def __init__(self, tfm):
        self.tfm = tfm

    def begin_batch(self):
        self.learner.xb = self.tfm(self.xb)


class ParamScheduler(Callback):
    _order = 1

    def __init__(self, pname, sched_funcs):
        self.pname, self.sched_funcs = pname, sched_funcs

    def begin_fit(self):
        if not isinstance(self.sched_funcs, (list, tuple)):
            self.sched_funcs = [self.sched_funcs] * len(self.opt.param_groups)

    def set_param(self):
        assert len(self.opt.param_groups) == len(self.sched_funcs)
        for pg, f in zip(self.opt.param_groups, self.sched_funcs):
            pg[self.pname] = f(self.n_epochs/self.epochs)

    def begin_batch(self):
        if self.state == 'train':
            self.set_param()


def annealer(f):
    def _inner(start, end):
        return partial(f, start, end)
    return _inner


@annealer
def sched_linear(start, end, pos):
    return start + pos*(end-start)


@annealer
def sched_cos(start, end, pos):
    return start + (1 + math.cos(math.pi*(1-pos))) * (end-start) / 2


@annealer
def sched_no(start, end, pos):  # pylint: disable=unused-argument
    return start


@annealer
def sched_exp(start, end, pos):
    return start * (end/start) ** pos


def cos_1cycle_anneal(start, high, end):
    return [sched_cos(start, high), sched_cos(high, end)]  # pylint: disable=no-value-for-parameter


def combine_scheds(pcts, scheds):
    assert sum(pcts) == 1.
    pcts = torch.tensor([0] + tbox.listify(pcts))
    assert torch.all(pcts >= 0)
    pcts = torch.cumsum(pcts, 0)

    def _inner(pos):
        idx = (pos >= pcts).nonzero().max()
        actual_pos = (pos-pcts[idx]) / (pcts[idx+1]-pcts[idx])
        return scheds[idx](actual_pos)
    return _inner


if __name__ == '__main__':
    a = torch.arange(0, 100)
    p = torch.linspace(0.01, 1, 100)

    # annealings = "NO LINEAR COS EXP".split()
    # fns = [sched_no, sched_lin, sched_cos, sched_exp]
    # for fn, t in zip(fns, annealings):
    #     f = fn(2, 1e-2)
    #     plt.plot(a, [f(o) for o in p], label=t)
    # plt.legend()
    # plt.show()

    sched = combine_scheds([0.3, 0.7], [sched_cos(0.3, 0.6), sched_cos(0.6, 0.2)])  # pylint: disable=no-value-for-parameter
    plt.plot(a, [sched(o) for o in p])
    plt.show()



import torch
from bijou.utils import rename

@rename('acc')
def accuracy(out, yb):
    return (torch.argmax(out, dim=1) == yb).float().mean()


@rename('acc')
def masked_accuracy(pred, target):
    _, pred = pred.max(dim=1)
    correct = pred[target.mask].eq(target.data[target.mask]).sum()
    acc = correct / target.mask.float().sum()
    return acc



"""
模型保存与加载
结果解解
各种计算指标
每层使用不同的学习率
各种优化技术：早停...
"""

import torch
from bijou.utils import ToolBox as tbox
import bijou.callbacks as cbks
from functools import partial
import matplotlib.pyplot as plt


class Learner():
    def __init__(self, model, opt, loss_func, data, metrics=None, cbs=None, cb_attrs=None, device=None):
        """
        Args:
            model: pytorch 模型
            opt: 优化器
            loss_func: 损失函数
            data: DataBunch
            metrics: 性能评价指标或评价指标列表
            cbs: callbacks对象列表
            cb_attrs: callbacks类列表，其中的每个callback会被作为 learner 的属性，
                以便于在fit后通过learner属性访问callback，例如Recorder
            device: cpu or gpu device
        """

        self.model, self.opt, self.data, self.loss_func = model, opt, data, loss_func
        self.state = 'train'  # 'train', 'val', 'test'
        self.infos = {'train': [], 'val': [], 'test': []}
        self.metric_values = {}

        cbs = tbox.listify(cbs)
        cb_attrs = tbox.listify(cb_attrs) + [cbks.Recorder]
        # if metrics:
        cb_attrs += [partial(cbks.AvgStatsCallback, metrics)]
        for cbf in cb_attrs:
            cb = cbf()
            setattr(self, cb.name, cb)
            cbs.append(cb)

        self.cbs = [cbks.TrainEvalCallback(),
                    cbks.CudaCallback(device),
                    cbks.ProgressBarCallback()] + cbs

    def __call__(self, cb_name, reverse=False):
        res = False
        cb_list = sorted(self.cbs, key=lambda x: x._order)  # pylint: disable=protected-access
        if reverse:
            cb_list = reversed(cb_list)
        for cb in cb_list:
            res = cb(cb_name) or res
        return res

    def one_batch(self, xb, yb):
        try:
            self.xb, self.yb = xb, yb
            self('begin_batch', reverse=False)
            self.pred = self.model(self.xb)
            self('after_pred', reverse=True)
            self.loss = self.loss_func(self.pred, self.yb)
            self('after_loss', reverse=True)
            if self.state != 'train':  # 若不是训练状态，则结束batch
                return
            self.loss.backward()
            self('after_backward', reverse=True)
            self.opt.step()
            self('after_step', reverse=True)
            self.opt.zero_grad()
        except cbks.CancelBatchException:
            self('after_cancel_batch', reverse=True)
        finally:
            self('after_batch', reverse=True)

    def all_batches(self, dl):
        self.iters = len(dl)
        try:
            for xb, yb in dl:
                self.one_batch(xb, yb)
        except cbks.CancelEpochException:
            self('after_cancel_epoch', reverse=True)

    def run_fit(self, epochs):
        self.epochs, self.loss = epochs, torch.tensor(0.)

        try:
            for cb in self.cbs:
                cb.set_learner(self)
            self('begin_fit', reverse=False)
            for epoch in range(epochs):
                self.epoch = epoch
                if not self('begin_epoch', reverse=False):
                    self.all_batches(self.data.train_dl)

                if self.data.valid_dl is not None:
                    with torch.no_grad():
                        if not self('begin_validate', reverse=False):
                            self.all_batches(self.data.valid_dl)
                self('after_epoch', reverse=True)

        except cbks.CancelTrainException:
            self('after_cancel_train', reverse=True)
        finally:
            self('after_fit', reverse=True)
            # self.model, self.opt, self.loss_func, self.data = None, None, None, None

    def test(self, test_dl):
        test_cb = cbks.TestCallback()
        test_cb.set_learner(self)
        self.cbs += [test_cb]
        if not self('begin_test', reverse=False):
            with torch.no_grad():
                self.all_batches(test_dl)
        self('after_test', reverse=True)
        self.cbs.remove(test_cb)

    def predict(self, data_tensor):
        self.pred_data_tensor = data_tensor
        for cb in self.cbs:
            cb.set_learner(self)
        self('begin_predict', reverse=False)
        with torch.no_grad():
            pred = self.model(self.pred_data_tensor)
        return pred

    def fit(self, epochs):
        self.run_fit(epochs)

    def fit_one_cycle(self, epochs, stage=(0.3, 0.7), start_lr=0.01, high_lr=0.5, end_lr=0.01, plot=True):
        sched = cbks.combine_scheds(stage,
                                    cbks.cos_1cycle_anneal(start_lr, high_lr, end_lr)
                                    )  # pylint: disable=no-value-for-parameter
        lr_shed_cb = cbks.ParamScheduler('lr', sched)
        self.cbs += [lr_shed_cb]
        self.run_fit(epochs)
        if plot:
            plt.figure(figsize=[10, 5])

            plt.subplot(121)
            self.recorder.plot_lr()
            plt.xlabel('n iter')
            plt.ylabel('learning rate')

            plt.subplot(122)
            self.recorder.plot_loss()
            plt.xlabel('n iter')
            plt.ylabel('loss')

        self.cbs.remove(lr_shed_cb)

    def find_lr(self, max_iter=100, min_lr=1e-6, max_lr=10):
        lr_find_cbk = cbks.LR_Find(max_iter, min_lr, max_lr)
        self.cbs += [lr_find_cbk]
        self.run_fit(max_iter)

        self.recorder.plot_lr_loss(skip_last=5)
        plt.xlabel('learning rate')
        plt.ylabel('loss')

        self.cbs.remove(lr_find_cbk)



import setuptools

with open("README.md", 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name="bijou",
    version="0.0.1",
    author="hitlic",
    author_email="liuchen.lic@gmail.com",
    description="",
    long_description=long_description,
    url="",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    
)